#include <stdlib.h> /* malloc */
#include <stdio.h>  /* printf */
#include <assert.h>
#include <sys/time.h>
#include <cuda.h>

/* Integer ceiling division */
#define DIV_CEIL(a,b) (a+b-1)/b
/* A time constant */
#define MICROSEC (1E-6)

int timeval_subtract (struct timeval *result, struct timeval *x, struct timeval *y);

/* Device code */
__global__ void non_atomic_inc(unsigned long int N, uint1 *value)
{
#define KERNEL_FLOP 0
#define KERNEL_MOP 2
	/* From nbody_kernel.cu, not really needed, good compilers exist */
	int i = __mul24(blockIdx.x,blockDim.x) + threadIdx.x;
	if (i < N) {
		value->x++;
	}
}

/* Host code */
int main(int argc, char** argv)
{
	unsigned long int N = 0;
	unsigned int threadsPerBlock = 0;
	unsigned int blocksPerGrid = 0;
	
	/* Handle input params */
	if (argc<3 || argc>4) {
		printf("Usage: %s <2^i threads> <2^j threadsPerBlock> [<2^k blocksPerGrid>]\n", argv[0]);
		exit(1);
	}
	assert(argc==3 || argc==4);
	int i;
	for(i=0; i<argc; i++) {
		assert(argv[i]);
	}
	N = 1 << atoi(argv[1]);
	threadsPerBlock = 1 << atoi(argv[2]);
	if (argc==4) {
		blocksPerGrid = 1 << atoi(argv[3]);
	} else {
		blocksPerGrid = DIV_CEIL(N,threadsPerBlock);
	}

	/* Hardware limits */
	assert(threadsPerBlock<=512);
	assert(blocksPerGrid<=65535);
	assert(blocksPerGrid*threadsPerBlock>=N); /* enough threads */

	/* Allocate vectors in device memory */
	size_t size = 1 * sizeof(uint1);
	assert(0<size);
	
	uint1 *h_x = NULL;
	uint1 *d_x = NULL;

	/* Alloc and set host and device memory */
	h_x = (uint1 *)malloc(size);
	assert(h_x);
	cudaMalloc((void **) &d_x, size);
	assert(d_x);
	*h_x = make_uint1(0);
	assert(h_x->x==0);
	
	/* Copy vector from host memory to device memory
	 * h_pts is an input vectors stored in host memory
	 */
	cudaMemcpy(d_x, h_x, size, cudaMemcpyHostToDevice);

	/* Start timer */
	double secs = 0.0;
	struct timeval start = {0L,0L}, end = {0L,0L}, elapsed = {0L,0L};
	assert(gettimeofday(&start,NULL)==0);

	/* Invoke kernel */
	non_atomic_inc <<<blocksPerGrid, threadsPerBlock>>> (N, d_x);
	cudaThreadSynchronize();

	/* Stop timer */
	assert(gettimeofday(&end,NULL)==0);
	assert(timeval_subtract(&elapsed, &end, &start)==0);
	secs = (double)elapsed.tv_sec + ((double)elapsed.tv_usec*MICROSEC);

	/* Copy result from device memory to host memory
	 * h_C contains the result in host memory
	 */
	cudaMemcpy(h_x, d_x, size, cudaMemcpyDeviceToHost);

	/* Check for correctness */
	assert(0 < h_x->x);

	/* Print results */

	printf("Secs: %f\n", secs);
	printf("GMops: %f\n", (N*sizeof(float)*KERNEL_MOP / secs)*1e-9);
	printf("GFlops: %f\n", (N*KERNEL_FLOP / secs)*1e-9);
	printf("X: %d\n", h_x->x);

	/* Free device/host memory */
	cudaFree(d_x); d_x = NULL;
	free(h_x); h_x = NULL;

	return 0;
}


/*
 * http://www.gnu.org/software/libtool/manual/libc/Elapsed-Time.html
 * Subtract the `struct timeval' values X and Y,
 * storing the result in RESULT.
 * Return 1 if the difference is negative, otherwise 0.
 */

int timeval_subtract (struct timeval *result, struct timeval *x, struct timeval *y) {
	/* Perform the carry for the later subtraction by updating y. */
	if (x->tv_usec < y->tv_usec) {
		int nsec = (y->tv_usec - x->tv_usec) / 1000000 + 1;
		y->tv_usec -= 1000000 * nsec;
		y->tv_sec += nsec;
	}
	if (x->tv_usec - y->tv_usec > 1000000) {
		int nsec = (x->tv_usec - y->tv_usec) / 1000000;
		y->tv_usec += 1000000 * nsec;
		y->tv_sec -= nsec;
	}

	/* Compute the time remaining to wait. tv_usec is certainly positive. */
	result->tv_sec = x->tv_sec - y->tv_sec;
	result->tv_usec = x->tv_usec - y->tv_usec;

	/* Return 1 if result is negative. */
	return x->tv_sec < y->tv_sec;
}

