/*
 *      simpson_cuda.cu
 *
 *      Copyright © 2010 Dionisio E Alonso <dealonso@gmail.com>
 *	Modifications by: Nicolás Wolovick <nicolasw@famaf.unc.edu.ar>
 *
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation, either version 3 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */


#include <cuda.h>
#include <stdlib.h> /* malloc */
#include <stdio.h>  /* printf */
#include <assert.h>
#include <sys/time.h> /* gettimeofday */
#include <math.h> /* M_PI and such */

/* 0 is for Dionisio, 1 is for Eze */
#define DEVICE 0
/* Block size, fully occupied */
#define BLOCK_SIZE 512

/* Integer ceiling division */
#define DIV_CEIL(a,b) (a+b-1)/b

/* Measure WallTime */
int timeval_subtract (struct timeval *result, struct timeval *x, struct timeval *y);
#define MICROSEC (1E-6)
#define WALLTIME(PROGRAM) ({double secs = 0.0; \
	struct timeval start = {0L,0L}, end = {0L,0L}, elapsed = {0L,0L}; \
	assert(gettimeofday(&start,NULL)==0); \
	PROGRAM; \
	assert(gettimeofday(&end,NULL)==0); \
	assert(timeval_subtract(&elapsed, &end, &start)==0);\
	secs = (double)elapsed.tv_sec + ((double)elapsed.tv_usec*MICROSEC);\
	printf("WallTime secs: %f\n",secs);\
	})


double simpson_rule_cpu(double a, double b, unsigned int intervals);
double simpson_rule_gpu(double a, double b, unsigned int intervals);

__global__ void simpson_kernel_fx(double a, unsigned int intervals, double interval_size, double *d_fx);
__global__ void simpson_kernel_sum(double a, unsigned int intervals, double interval_size, double *d_fx, double *d_simpson);
__host__ __device__ double f(double x) { return sin(x); }


int main(int argc, char **argv)
{
	const double a=0.0, b=M_PI;
	unsigned int intervals=1E6;
	double cpu=0.0, gpu=0.0;
	WALLTIME(cpu = simpson_rule_cpu(a,b,intervals);
	);
	WALLTIME(gpu = simpson_rule_gpu(a,b,intervals);
		assert(cudaThreadSynchronize()==cudaSuccess);
	);
	printf("%.10f, %.10f\n", cpu, gpu);
	assert(cpu==gpu);
	
	return 0;
}


double simpson_rule_gpu(double a, double b, unsigned int intervals)
{
	double interval_size=0.0;
	double *d_simpson=NULL, *d_fx=NULL;
	double h_simpson=0.0;

	interval_size = (b-a)/intervals;

	/* Set device to use */
	cudaSetDevice(DEVICE);

	cudaMalloc((void **) &d_fx, (intervals+1)*sizeof(double));
	cudaMalloc((void **) &d_simpson, sizeof(double));
	cudaMemset((void *) d_simpson, 0, sizeof(double));

	/* Invoke kernel */
	dim3 dim_block(BLOCK_SIZE);
	dim3 dim_grid(DIV_CEIL(intervals, dim_block.x));

	simpson_kernel_fx<<<dim_grid, dim_block>>>(a, intervals, interval_size, d_fx);
	simpson_kernel_sum<<<dim_grid, dim_block>>>(a, intervals, interval_size, d_fx, d_simpson);
	/* End kernel invocation */

	/* Copy results back */
	cudaMemcpy((void *)&h_simpson, (void *)d_simpson, sizeof(double), cudaMemcpyDeviceToHost);

	/* Free device memory */
	cudaFree(d_fx);
	cudaFree(d_simpson);

	return interval_size/3*h_simpson;
}


__global__ void simpson_kernel_fx(double a, unsigned int intervals, double interval_size, double *d_fx)
{
	const unsigned int i = blockIdx.x * blockDim.x + threadIdx.x;

	if (i<=intervals) {
		if (i==0 || i==intervals)
			d_fx[i] = f(a+i*interval_size);
		else if (i%2 == 1)
			d_fx[i] =  f(a+i*interval_size)*4;
		else
			d_fx[i] = f(a+i*interval_size)*2;
	}
}


__global__ void simpson_kernel_sum(double a, unsigned int intervals, double interval_size, double *d_fx, double *d_simpson)
{
	const unsigned int i = blockIdx.x * blockDim.x + threadIdx.x;

	if (i<=intervals) {
		*d_simpson = *d_simpson + d_fx[i];
	}
}


double simpson_rule_cpu(double a, double b, unsigned int intervals)
{
	double interval_size=0.0;
	double result=0.0;
	unsigned int i=0;

	interval_size = (b-a)/intervals;
	for(i=0; i<=intervals; i++) {
		if (i==0 || i==intervals)
			result += f(a+i*interval_size);
		else if (i%2 == 1)
			result += f(a+i*interval_size)*4;
		else
			result += f(a+i*interval_size)*2;
	}
	result = interval_size/3*result;

	return result;
}

/*
 * http://www.gnu.org/software/libtool/manual/libc/Elapsed-Time.html
 * Subtract the `struct timeval' values X and Y,
 * storing the result in RESULT.
 * Return 1 if the difference is negative, otherwise 0.
 */

int timeval_subtract (struct timeval *result, struct timeval *x, struct timeval *y) {
	/* Perform the carry for the later subtraction by updating y. */
	if (x->tv_usec < y->tv_usec) {
		int nsec = (y->tv_usec - x->tv_usec) / 1000000 + 1;
		y->tv_usec -= 1000000 * nsec;
		y->tv_sec += nsec;
	}
	if (x->tv_usec - y->tv_usec > 1000000) {
		int nsec = (x->tv_usec - y->tv_usec) / 1000000;
		y->tv_usec += 1000000 * nsec;
		y->tv_sec -= nsec;
	}

	/* Compute the time remaining to wait. tv_usec is certainly positive. */
	result->tv_sec = x->tv_sec - y->tv_sec;
	result->tv_usec = x->tv_usec - y->tv_usec;

	/* Return 1 if result is negative. */
	return x->tv_sec < y->tv_sec;
}
